# speed of light (m s^-1)
c = 2.99792458e8

# little g (m s^-2)
g = 9.806651

# Planck's constant (kg m^2 s)
h = 6.62607015e-34

# Reduced Planck's constant (kg m^2 s)
hbar = h / 6.283185307179586

# Boltzmann's constant (kg m^2 s^-2 K^-1)
k_B = 1.380649e-23

# CMB temperature (K)
T_CMB = 2.72548

# equatorial radius of the earth (m)
earth_radius = 6.378137e6

MARIA_MIN_NU_HZ = 1e6  # 1 MHz
MARIA_MAX_NU_HZ = 15e12  # 150 THz

DEFAULT_TIME_FORMAT = "YYYY-MM-DD HH:mm:ss.SSS ZZ"
DEFAULT_BAR_FORMAT = "{l_bar}{bar:16}{r_bar}"
